
import javax.swing.*;
import java.awt.geom.*;

/** The naval vehicles' superclass.  Subs, torpedoes, surface ships, etc.
    inherit from this class.  The class defines the main stats fore
    each vehicle as well as methods to access and mutate them.
*/
class Vehicle implements Runnable {
    private Thread runner;
    private String ownership;
    private double depth;
    private double speed;
    private double bearing;
    // relative to the established origin
    private Point2D.Double coord = new Point2D.Double();
    // at vehicle's coords
    private double seaDepth = 0; // will determine with first movement() call
    private double intendedDepth = 0;
    private double intendedSpeed = 0;
    private double intendedBearing = 0;

    /** @param aOwnership vehicle's political allegiance
	@param aDepth depth in feet
	@param aSpeed speed in knots
	@param aBearing bearing in compass degrees
	@param aCoord coordinates relative in nautical miles to the set origin
    */
    Vehicle(String aOwnership, double aDepth, double aSpeed, double aBearing, Point2D.Double aCoord) {
	ownership = aOwnership;
	depth = aDepth;
	speed = aSpeed;
	bearing = aBearing;
	coord = aCoord;
	if (runner == null) {
	    runner = new Thread(this);
	    runner.start();
	}
    }

    public void run() {
	Thread thisThread = Thread.currentThread();
	while (runner == thisThread) {
	    movement();
	    try {
		Thread.sleep(100);
	    } catch (InterruptedException e) { }
	}
    }

    void movement() {
	adjustDepth();
	adjustSpeed();
	adjustBearing();
	adjustPosition();
	setSeaDepth();
	checkCollision();
    }

    /* Accessors
     */
    Point2D.Double getCoord() {
	return (Point2D.Double)coord.clone();
    }

    // Strings are constants and can therefore be shared (cp J2SE API)
    String getOwnership() {
	return ownership;
    }

    double getDepth() {
	return depth;
    }

    double getSpeed() {
	return speed;
    }

    double getBearing() {
	return bearing;
    }

    double getSeaDepth() {
	return seaDepth;
    }

    /* Mutators
     */
    void setIntendedDepth(double d) {
	intendedDepth = d;
    }

    void setIntendedSpeed(double d) {
	intendedSpeed = d;
    }

    void setIntendedBearing(double d) {
	intendedBearing = d;
    }

    void setRunner(Thread t) {
	runner = t;
    }





    void adjustDepth() {
	double depthChange = (intendedDepth - depth) * Benel.DT;

	depth += depthChange;
    }

    void adjustSpeed() {
	double speedChange = (intendedSpeed - speed) * Benel.DT;

	speed += speedChange;
    }

    void adjustBearing() {
	double bearingChangeCheck = intendedBearing - bearing;
	double bearingChange;

	if (Math.abs(bearingChangeCheck) > 180) {
	    if (intendedBearing < bearing)
		bearingChange = Benel.DT * (360 - bearing + intendedBearing);
	    else
		bearingChange = - Benel.DT * (360 - intendedBearing + bearing);
	}
	else
	    /* 20 serves as baseline maneuverability:
	       subs going faster can change direction quicker 
	       since passing water exerts greater force; 
	       eventually need more sophisticated model */
	    bearingChange = Benel.DT * bearingChangeCheck * speed / 20; 
	bearing += bearingChange;
	if (bearing < 0)
	    bearing += 360;
	else if (bearing > 360)
	    bearing -= 360;
    }

    /** Incrementally adjusts vehicle's coordinates based on
	bearing and speed.
    */
    void adjustPosition() {
	// adjusts to rad and calculates movt during DT
	// bearing relative to due North, increasing clockwise
	coord.x += (double)speed * Math.sin((double)(Math.PI / 180 * bearing)) * Benel.DT * Benel.DP;
	coord.y += (double)speed * Math.cos((double)(Math.PI / 180 * bearing)) * Benel.DT * Benel.DP;
    }

    /** Determines the depth of the sea at the vehicle's
	coordinates. Based on htFld.bmp, where dark, or little
	blue, is deep, and light, bright blue is high.
    */
    void setSeaDepth() {
	Point2D.Double origin = Benel.getOrigin();
	Point2D.Double coordRecen = new Point2D.Double();
	int[] htFld = Benel.getHtFld();

	coordRecen.x = origin.x + coord.x;
	coordRecen.y = origin.y - coord.y;

	int n = (int)(320 * coordRecen.y + coordRecen.x);
	// check to make sure don't try to read off map
	if (n >= htFld.length)
	    n = htFld.length - 1;
	else if (n < 0)
	    n = 0;
	seaDepth = 10 * (255 - htFld[n]); // may want to alter depth factor later
    }

    void checkCollision() {
	if (depth >= seaDepth) {
 	    setRunner(null);
	    Benel.setOwn(null);
	    JOptionPane.showMessageDialog(null, "Collision with the sea floor!");
	}
    }    
}






class Sub extends Vehicle {

    Sub(String aOwnership, double aDepth, double aSpeed, double aBearing, Point2D.Double aCoord) {
	super(aOwnership, aDepth, aSpeed, aBearing, aCoord);
    }
}





/*
class OwnSub extends Sub {

    void OwnSub() {
    }

    void ownSubMovement() {
    }
}
*/





/*
class ComputerSub extends Sub {
}
*/





class Torpedo extends Vehicle {

    Torpedo(String aOwnership, double aDepth, double aSpeed, double aBearing, Point2D.Double aCoord) {
	super(aOwnership, aDepth, aSpeed, aBearing, aCoord);
    }
}
